/*
 *  display.c
 *
 *  Copyright 2007-2010 Jon Robey, Rachel Robey, and Bob Robey, brobey@lanl.gov 
 *  All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 *
 */

#include "display.h"

#include "config.h"

#ifdef USE_MPI
#include <mpi.h>
#else
#define MPE_NOMPI
#define CLOG_NOMPI
#endif
#define MPE_INTERNAL
#define MPE_GRAPHICS
#include "mpe.h"
#define MPE_DRAG_FIXED_RECT 8
#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
//keyboard input files
#include <termios.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <math.h>

#define MIN(a,b) ((a)>(b)?(b):(a))
#define MAX(a,b) ((a)<(b)?(b):(a))

// Control flags
static struct {
   unsigned int displayflag : 1;
} display_flags={1};

#define TEXTHEIGHT 15
#define SCALEWIDTH 75
#define SCALETEXT   3

static unsigned int width     = 1000;
static unsigned int height    = 600;
static unsigned int conHeight = 400;
static unsigned int conWidth  = 925;
static unsigned int graHeight = 170;
static unsigned int graWidth  = 925;

static double xmin = 0.0;
static double xmax = 0.0;
static double ymin = 0.0;
static double ymax = 0.0;
static double xconv = 0.0;
static double yconv = 0.0;
static double colScale = 10.0;
static double graScale = 11.0;
static int graRow = 10;
static int i_cell = -1;
static int j_cell = -1;

static MPE_XGraph graph;
static MPE_Color *color_array;
static unsigned int ncolors=256;
static char *label=NULL;
static XFontStruct *font_info;

static int DrawDragVisual(int dragVisual, MPE_XGraph graph, int x1, int y_1, int x2, int y_2, double ratio );
static int ConvertCoords(int dragVisual, int x1, int y_1, int x2, int y_2, double ratio,
                         int *cx1, int *cy_1, int *cwidth, int *cheight );
static void draw_contour(unsigned int imax, unsigned int jmax, unsigned int my_offset, unsigned int mysize,
                         double **temp,double *x, double *dx, double *y, double *dy);
static void draw_graph(unsigned int imax, unsigned int jmax, unsigned int my_offset, unsigned int mysize,
           double **temp,double *x, double *dx);


void display_init(int *argc, char ***argv, unsigned int iwidth, unsigned int iheight, unsigned int icontourheight,
                  double xmin_in, double xmax_in, double ymin_in, double ymax_in){
    
   int ierr;
   char fontname[30];
   int rank=0;
   char *displayname=NULL;

   /* Open the graphics display */
   
   if (*argc > 2 && strcmp( *argv[1], "-display" ) == 0) {
      displayname = (char *)malloc( strlen( *argv[2] ) + 1 );
      strcpy(displayname, *argv[2]);
   }

   if (displayname == NULL){
      displayname = getenv("DISPLAY");
      if (displayname == NULL){
         displayname = (char *)malloc(8);
         displayname = ":0";
      }
   }
    
   width  = iwidth;
   height = iheight;

   conHeight = icontourheight;
   conWidth  = width - SCALEWIDTH;

   graHeight = height - conHeight - 2*TEXTHEIGHT;
   if (graHeight < 0) graHeight  = 0;
   graWidth  = width - SCALEWIDTH;
   
   xmin = xmin_in;
   xmax = xmax_in;
   ymin = ymin_in;
   ymax = ymax_in;
    
   xconv = (double)conWidth/ (xmax-xmin);
   yconv = (double)conHeight/(ymax-ymin);
   
#ifdef USE_MPI
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);
   if (MPE_Open_graphics( &graph, MPI_COMM_WORLD, displayname,
                         -1, -1, width, height, 0) != MPE_SUCCESS){
#else
   if (MPE_Open_graphics( &graph, 0, displayname,
                          -1, -1, width, height, 0) != MPE_SUCCESS){
#endif
      printf("Problem opening graphics window\n");
      exit(-1);
   }
        
   if (rank == 0){
      XSelectInput( graph->xwin->disp, graph->xwin->win, MPE_XEVT_IDLE_MASK |
                  ButtonPressMask   |
                  ButtonReleaseMask |
                  Button1MotionMask |
                  KeyPressMask      |
                  ExposureMask      |
                  StructureNotifyMask);
   }
        
   MPE_Update(graph);
        
   // if we want to change fonts, we can load a new one with the following
   strcpy(fontname,"fixed");
   font_info = XLoadQueryFont( graph->xwin->disp, fontname );
    
   color_array = (MPE_Color *) malloc(sizeof(MPE_Color)*ncolors);
    
   ierr = MPE_Make_color_array(graph, ncolors, color_array);
   if (ierr && rank == 0) printf("Error(Make_color_array): ierr is %d\n",ierr);
}
    
    
void display_close(void){
   MPE_Close_graphics(&graph);
}
    
void display_set_label(char *text)
{
   label=text;
}
   
void display_data_interactive(unsigned int imax, unsigned int jmax, unsigned int my_offset, unsigned int mysize,
                              double **temp,double *x, double *dx, double *y, double *dy,
                              int *button, char *key, int *xloc, int *yloc, double *xcor, double *ycor,
                              int *icell, int *jcell, int *special_event)
{
   XEvent event;
   int pos_x, pos_y;
   unsigned int keys_buttons;
   Window root, child;
   int root_x, root_y;
   int drag_event;
   int i, j;
   
   display_data(imax, jmax, my_offset, mysize, temp, x, dx, y, dy);
   //sleep(WAIT_TIME);
   //sleep(1);
   
   int loop = 1;
   while (loop) {
      loop = 0;
      drag_event = 0;
      if (*button != -1 || key != '\0') loop = 1;
      display_get_event(button, key, xloc, yloc, xcor, ycor, special_event);
      if (*button != -1){
         if (*button == 1){
            while (1) {

               while (XCheckTypedEvent(graph->xwin->disp, ButtonPress, &event));
               /* Wait for release; if on correct button, exit */
               XMaskEvent(graph->xwin->disp, Button1MotionMask | ButtonReleaseMask | PointerMotionHintMask, &event);
               if (event.type == MotionNotify ){
                  //printf("DEBUG -- motion notify\n");
                  // We throw away the next motion commands
                  while (XCheckMaskEvent(graph->xwin->disp, Button1MotionMask, &event));
                  // And query the pointer directly -- this makes the process much more responsive
                  XQueryPointer(graph->xwin->disp, event.xmotion.window, &root, &child, &root_x, &root_y,
                                      &pos_x, &pos_y, &keys_buttons);
                  //printf("DEBUG -- %d %d\n",pos_x,pos_y);
                  if (pos_y < 0) pos_y = 0; 
                  if (pos_x < 0) pos_x = 0; 
                  if (pos_y >= conHeight) pos_y = conHeight-1; 
                  if (pos_x >= conWidth)  pos_x = conWidth-1; 
                  //printf("DEBUG result -- %d %d\n",pos_x,pos_y);
                  // If motion is smaller than 5 pixels, treat it as a single mouse click
                  // Drawing the contour to get rid of the previous drag visuals
                  if (drag_event){
                     if (abs(pos_x - *xloc) <= 5 || abs(pos_y - *yloc) <= 5) drag_event = 0;
                     draw_contour(imax, jmax, my_offset, mysize, temp, x, dx, y, dy);
                  }
                  if (! drag_event){
                     if (abs(pos_x - *xloc) > 5 || abs(pos_y - *yloc) > 5) drag_event = 1;
                  }
                  if (drag_event){
                     DrawDragVisual(MPE_DRAG_RECT, graph, *xloc, *yloc, pos_x, pos_y, 1.0);
                     //DrawDragVisual(MPE_DRAG_RECT, graph, *xloc, *yloc, event.xmotion.x, event.xmotion.y, 1.0);
                  }
               }
               if (event.type == ButtonRelease) {
                  //printf("DEBUG -- button release on button %d needs to match %d\n",event.xbutton.button, ibutton);
                  if (event.xbutton.button == *button) {
                     pos_x = event.xbutton.x;
                     pos_y = event.xbutton.y;
                     if (pos_y < 0) pos_y = 0; 
                     if (pos_x < 0) pos_x = 0; 
                     if (pos_y >= conHeight) pos_y = conHeight-1; 
                     if (pos_x >= conWidth)  pos_x = conWidth-1; 
                     if (drag_event) {
                        double xpos1 = xmin + (double)*xloc/xconv;
                        double ypos1 = ymin + (double)(conHeight - *yloc - 1)/yconv;
                        double xpos2 = xmin + (double)pos_x/xconv;
                        double ypos2 = ymin + (double)(conHeight - pos_y - 1)/yconv;
                    
                        //printf("Got here %d %d %lf %lf\n",pos_x,pos_y,pos_x/xconv,pos_y/yconv);
                        double new_xmin = MIN(xpos1, xpos2);
                        double new_ymin = MIN(ypos1, ypos2);
                        double new_xmax = MAX(xpos1, xpos2);
                        double new_ymax = MAX(ypos1, ypos2);
                        printf(" New window is xmin %lf xmax %lf ymin %lf ymax %lf\n",new_xmin, new_xmax, new_ymin, new_ymax);
                        display_set_window(new_xmin, new_xmax, new_ymin, new_ymax);
                        //XMaskEvent(graph->xwin->disp, ButtonPress, &event);
                        //XMaskEvent(graph->xwin->disp, Button1MotionMask | ButtonReleaseMask, &event);
                        *button = -1;
                     }
                     break;
                  }
               }
            }
            *xcor=xmin+(double)*xloc/xconv;
            *ycor=ymin+(double)(conHeight-*yloc-1)/yconv;
            *jcell = -1;
            j_cell = -1;
            for(j=0; j<jmax; j++){
               //printf("DEBUG %d %lf %lf %lf\n",j,*ycor,y[j]-0.5*dy[j],y[j]+0.5*dy[j]);
               if (*ycor >= y[j]-0.5*dy[j] && *ycor < y[j]+0.5*dy[j]){
                  //printf("DEBUG Match is %d %d\n",j,jmax);
                  *jcell = j;
                  j_cell = j;
                  break;
               } 
            }
            *icell = -1;
            i_cell = -1;
            for(i=0; i<imax; i++){
               if (*xcor >= x[i]-0.5*dx[i] && *xcor < x[i]+0.5*dx[i]){
                  *icell = i;
                  i_cell = i;
                  break;
               } 
            }
            //printf("INPUT -- button is %d at location %d %d %lf %lf cell index %d %d\n", *button, *xloc, *yloc, *xcor, *ycor, i_cell, j_cell);
            loop = 0;
         }
         
      }
      if (*key == 'g') loop = 0;
      //if (ibutton || key !='\0' || special_event || icell >=0 ) {
         //printf("INPUT -- button is %d key is '%c' special_event is %d icell is %d loc is %d %d coor is %lf %lf\n",
         //      ibutton,      *key,        *special_event,      *icell,      *xloc, *yloc, *xcor, *ycor);
         //printf("INPUT -- button is %d key is '%c' special_event is %d\n", ibutton, *key, *special_event);
                
                // icell is %d loc is %d %d coor is %lf %lf\n",
                //      *icell,      *xloc, *yloc, *xcor, *ycor);
      //}
      //i++;
      //if (i>10000000) {
         loop = 0;
      //}
      
   }
   
   //printf("DEBUG \n");
}
      
void display_data(unsigned int imax, unsigned int jmax, unsigned int my_offset, unsigned int mysize,
                  double **temp,double *x, double *dx, double *y, double *dy)
{
   unsigned int plot_value;
   char maxGra[20], maxCol[20], min[20];
   char string[80];
   
   if (! display_flags.displayflag) return;
   
   draw_contour(imax, jmax, my_offset, mysize, temp, x, dx, y, dy);

   draw_graph(imax, jmax, my_offset, mysize, temp, x, dx);
   
   if(my_offset==0){
      //graph scale
      MPE_Fill_rectangle( graph, graWidth, conHeight, SCALEWIDTH, graHeight-(graHeight*(colScale/graScale)), MPE_BLACK);
      double scaleHeight=graHeight*(colScale/graScale);
      for (unsigned int i=0; i<=scaleHeight; i++) {
      //draw color for scale
        plot_value = ncolors - ((double)ncolors*i/scaleHeight) - 1;
         //printf("plot_value %d %d %lf %lf %d\n", ncolors, i, scaleHeight, (double)ncolors*i/scaleHeight, plot_value);
              MPE_Fill_rectangle( graph, graWidth, height-2*TEXTHEIGHT-i, SCALEWIDTH, 1, color_array[plot_value]);
      }
      //graph legend
      sprintf( maxGra, "%6.4e", graScale);
      sprintf( maxCol, "%6.4e", colScale);
      sprintf( min, "%6.4e", 0.0);
      MPE_Fill_rectangle(graph, 0, height-2*TEXTHEIGHT, graWidth+SCALEWIDTH, 2*TEXTHEIGHT, MPE_WHITE);
      MPE_Fill_rectangle( graph, graWidth, 0, 2, height, MPE_BLACK);
      MPE_Fill_rectangle(graph, 0, height-2*TEXTHEIGHT, width, 2, MPE_BLACK);
      if (i_cell > -1 && j_cell > -1){
         sprintf(string,"Selected Cell is %d %d with value %lf", i_cell, j_cell, temp[j_cell][i_cell]);
         MPE_Draw_string( graph, 3, height-TEXTHEIGHT-3, MPE_BLACK, string);
      } else {
         MPE_Draw_string( graph, 3, height-TEXTHEIGHT-3, color_array[ncolors/2], "Color Scale upper bound, graph row");
      }
      MPE_Draw_string( graph, graWidth+5,conHeight-3,MPE_BLACK,maxGra);
      if(graHeight-(graHeight*(colScale/graScale))>TEXTHEIGHT) MPE_Draw_string( graph , graWidth+5, height-2*TEXTHEIGHT-(graHeight*(colScale/graScale))-3, MPE_WHITE, maxCol);
      MPE_Draw_string( graph, graWidth+5,height-TEXTHEIGHT-3,MPE_BLACK,min);
      //Label
      MPE_Fill_rectangle(graph, 0, height-TEXTHEIGHT, graWidth, 2, MPE_BLACK);
      if(label!=NULL)MPE_Draw_string( graph, 3, height-3, MPE_BLACK, label);
   }
   MPE_Update( graph );
}
  
static void draw_contour(unsigned int imax, unsigned int jmax, unsigned int my_offset, unsigned int mysize,
                           double **temp,double *x, double *dx, double *y, double *dy){
   unsigned int plot_value;
   int yloc, ywid, xloc, xwid;
   int xloc1, xloc2, yloc1, yloc2;
   unsigned int jstart, jend, istart, iend;
   unsigned int i, j;
   char string[80];
   
   for(j=1; j<mysize && y[j] < ymin; j++);
   jstart = j-1;
   for(j=mysize-1; j>=1 && y[j] > ymax; j--);
   jend = j+1;
   for(i=1; i<imax && x[i] < xmin; i++);
   istart = i-1;
   for(i=imax-1; i>=1 && x[i] > xmax; i--);
   iend = i+1;
   
   //   if (x[j][0] > xmin) break;
   //}
   //jend = j;
   //for(unsigned int j=0; j<mysize; j++){
   //printf("jstart is %d jend is %d\n",jstart,jend);
      
   for(j=jstart; j<jend; j++){
      yloc = (int)((y[j]-0.5*dy[j]-ymin)*yconv        + 0.5);
      ywid = (int)((y[j]+0.5*dy[j]-ymin)*yconv - yloc + 0.5);
      yloc = conHeight - yloc -1;
      
      if (ywid <= 0) continue;

      //if (yloc + ywid < 0) continue;
      if (yloc >= conHeight) continue;
      
      for(i=istart; i<iend; i++){
         xloc = (int)((x[i]-0.5*dx[i]-xmin)*xconv        + 0.5);
         xwid = (int)((x[i]+0.5*dx[i]-xmin)*xconv - xloc + 0.5);
         if (xwid <= 0) continue;
         if (xloc >= conWidth || xloc + xwid < 0) continue;
         
         if (xloc + xwid > conWidth) xwid = conWidth - xloc;
         if (xloc < 0) xloc = 0;
         
         plot_value = ncolors - ((double)ncolors*temp[j][i]/colScale) + 2;
         if (plot_value < 2) plot_value = 2;
         if (plot_value >= ncolors) plot_value = ncolors - 1;
         if(isnan(temp[j][i]))MPE_Fill_rectangle( graph, xloc, yloc, xwid, ywid, MPE_WHITE);
         else if(temp[j][i]<0)MPE_Fill_rectangle( graph, xloc, yloc, xwid, ywid, MPE_WHITE);
         else if(temp[j][i]>colScale)MPE_Fill_rectangle( graph, xloc, yloc, xwid, ywid, MPE_BLACK);
         else MPE_Fill_rectangle(graph, xloc, yloc, xwid, ywid, color_array[plot_value]);
      }
      if(graRow-my_offset==j){
         MPE_Fill_rectangle(graph,width+2,yloc,SCALEWIDTH,(ywid>1)?ywid:1,color_array[ncolors/2]);
      }
   }

   //printf("INPUT -- cell index %d %d\n", i_cell, j_cell);

   MPE_Fill_rectangle(graph, conHeight, conWidth, 0, 0, MPE_BLACK);
   if (i_cell > -1 && j_cell > -1){
      xloc1 = (int)((x[i_cell]-0.5*dx[i_cell]-xmin)*xconv + 0.5);
      xloc2 = (int)((x[i_cell]+0.5*dx[i_cell]-xmin)*xconv + 0.5);
      yloc1 = (int)((y[j_cell]-0.5*dy[j_cell]-ymin)*yconv + 0.5);
      yloc2 = (int)((y[j_cell]+0.5*dy[j_cell]-ymin)*yconv + 0.5);
      yloc1 = conHeight - yloc1 - 1;
      yloc2 = conHeight - yloc2 - 1;
      //printf("INPUT -- location %d %d %d %d cell index %d %d\n", xloc1, xloc2, yloc1, yloc2, i_cell, j_cell);
      
      MPE_Draw_line(graph, xloc1, yloc2, xloc2, yloc2, MPE_BLACK);
      MPE_Draw_line(graph, xloc1, yloc1, xloc2, yloc1, MPE_BLACK);
      MPE_Draw_line(graph, xloc1, yloc1, xloc1, yloc2, MPE_BLACK);
      MPE_Draw_line(graph, xloc2, yloc1, xloc2, yloc2, MPE_BLACK);

      MPE_Fill_rectangle(graph, 0, height-TEXTHEIGHT+1, width, height-1, MPE_WHITE);
      sprintf(string,"Selected Cell is %d %d with value %lf", i_cell, j_cell, temp[j_cell][i_cell]);
      MPE_Draw_string( graph, 3, height-TEXTHEIGHT-3, MPE_BLACK, string);
   } else {
      MPE_Draw_string( graph, 3, height-TEXTHEIGHT-3, color_array[ncolors/2], "Color Scale upper bound, graph row");
   }
}

static void draw_graph(unsigned int imax, unsigned int jmax, unsigned int my_offset, unsigned int mysize,
              double **temp,double *x, double *dx){
   int xloc, xwid;
      
   //draw graph
   if (my_offset<=graRow && my_offset+mysize>graRow){
      MPE_Fill_rectangle(graph, 0, conHeight, graWidth, graHeight, MPE_WHITE);
      MPE_Fill_rectangle(graph, 0, conHeight, graWidth, 2, MPE_BLACK);
      MPE_Fill_rectangle( graph, 0, (double)height-(double)2*TEXTHEIGHT-(double)graHeight*(colScale/graScale), graWidth, 1, color_array[ncolors/2]);
      for(unsigned int i=0; i<imax; i++){
         xloc = (int)((x[i]-0.5*dx[i]-xmin)*xconv        + 0.5);
         xwid = (int)((x[i]+0.5*dx[i]-xmin)*xconv - xloc + 0.5);
         if (xwid <= 0) continue;
         if (xloc >= conWidth || xloc + xwid < 0) continue;
         
         if (xloc + xwid > conWidth) xwid = conWidth - xloc;
         if (xloc < 0) xloc = 0;
         MPE_Fill_rectangle( graph, xloc, (double)height-(double)2*TEXTHEIGHT-(double)graHeight*(temp[graRow-my_offset][i]/graScale), xwid, 1, color_array[1]);
      }
   }
}
   
/********************************************************************************/
void display_get_event(int *button, char *key, int *xloc, int *yloc, double *xcor, double *ycor,
                       int *special_event)
/********************************************************************************/
{
   //double xmid, ymid;
   //int xrel, yrel;
   XEvent event;
   long EventMask;
   int EventFlag;
   char keys[20];
   int numChar;
   KeySym keysym;
   XComposeStatus compose;
   int rank=0;
        
#ifdef USE_MPI
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif
        
   if (graph->Cookie != MPE_G_COOKIE) {
      printf("Handle argument is incorrect or corrupted\n" );
   }
        
   *key = '\0';
   *button = -1;
   *xcor = 0.0;
   *ycor = 0.0;
   *special_event = 0;
        
   EventMask= ButtonPressMask | ButtonReleaseMask | KeyPressMask | ExposureMask | StructureNotifyMask ;

   
   EventFlag=0;
   //if (rank ==0){
      if(XCheckWindowEvent(graph->xwin->disp,graph->xwin->win,EventMask,&event)){
         EventFlag=1;
      }
   //}
#ifdef USE_MPI
   MPI_Bcast(&EventFlag, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
        
   if (EventFlag){
#ifdef USE_MPI
      MPI_Bcast(&event, sizeof(XEvent), MPI_BYTE, 0, MPI_COMM_WORLD);
#endif
            
      //printf("Event type is %d\n",event.type);
      switch(event.type){
         case ButtonPress:
            //printf("Button Press is %d\n",event.xbutton.button);
            *button=event.xbutton.button;
            *xloc=event.xbutton.x;
            *yloc=event.xbutton.y;
            *xcor=xmin+(double)*xloc/xconv;
            *ycor=ymin+(double)(height-*yloc -1)/yconv;
                            
            //printf("DEBUG graphics -- button is %d loc is %d %d cor is %lf %lf\n",
            //                       *button,    *xloc,*yloc,*xcor,*ycor);
            break;
         case KeyPress:
            if (rank == 0) {
               numChar=XLookupString((XKeyEvent *) &event, keys, 20, &keysym, &compose);
               printf("%c:%c:%d:\n",keys[0],keys[1],(int)keysym);
               if(((keysym>=XK_KP_Space)&&(keysym<=XK_KP_9))
                  ||((keysym>XK_space)&&(keysym<XK_asciitilde))){
                  *key=keys[0];
               }
            }
#ifdef USE_MPI
            MPI_Bcast(key,           1, MPI_CHAR,   0, MPI_COMM_WORLD);
#endif
                    
            *xloc=event.xkey.x;
            *yloc=event.xkey.y;
               
            *xcor=xmin+(double)*xloc/xconv;
            *ycor=ymin+(double)(height-*yloc-1)/yconv;
                    
            //printf("DEBUG graphics -- key is '%c' loc is %d %d cor is %lf %lf\n",
            //                         *key,      *xloc,*yloc,*xcor,*ycor);
            break;
         case Expose:
            if (event.xexpose.count == 0){
               *special_event=1;
            }
            //printf("DEBUG graphics -- special event is \n",*special_event);
            break;
         case ConfigureNotify:
            /* Window has been resized */
            width=event.xconfigure.width; //-SCALEWIDTH;
            double ratio = (double)event.xconfigure.height/(double)height;
            height *= ratio;
            
            conHeight *= ratio;
            conWidth  = width - SCALEWIDTH;
            
            graHeight = height - conHeight - 2*TEXTHEIGHT;
            if (graHeight < 0) graHeight  = 0;
            graWidth  = width - SCALEWIDTH;
            
            xconv = (double)conWidth/ (xmax-xmin);
            yconv = (double)conHeight/(ymax-ymin);

            //printf("DEBUG graphics -- window has been resized to %d %d\n",width, height);
            break;
      } // switch
         
      //printf("DEBUG -- button is %d key is '%c' loc is %d %d cor is %lf %lf\n",
      //                *button,     *key,      *xloc,*yloc,*xcor,*ycor);
   }
      
   //MPE_Update(graph);
   //usleep(300000);
        
   return;
}
    
    
void display_on(void){
   display_flags.displayflag = 1;
}

void display_off(void){
   display_flags.displayflag = 0;
}
   
/* ratio is in terms of height/width */
   
static int DrawDragVisual(int dragVisual, MPE_XGraph graph, int x1, int y_1, int x2, int y_2, double ratio )
{
   int width, height, top, left;
      
   ConvertCoords( dragVisual, x1, y_1, x2, y_2, ratio, &left, &top,
                  &width, &height );
      
   switch (dragVisual) {
      case MPE_DRAG_NONE: break;
      case MPE_DRAG_FIXED_RECT:
      case MPE_DRAG_RECT:
         XDrawRectangle( graph->xwin->disp, graph->xwin->win,
                        graph->xwin->gc.set, left, top, width, height );
         break;
      default:
      case MPE_DRAG_LINE:
         XDrawLine( graph->xwin->disp, graph->xwin->win,
                     graph->xwin->gc.set, left, top, width, height );
         break;
            
   }
   MPE_Update( graph );
   return 0;
}
   
/* convert start and end coordinates to x,y,width,height coords. */
/* different transformations required for each dragVisual. */
/* y_1 instead of y1 because y1 is a function in math.h (Bessel function
   of the second kind) */
static int ConvertCoords(int dragVisual, int x1, int y_1, int x2, int y_2, double ratio,
                           int *cx1, int *cy_1, int *cwidth, int *cheight )
{
   int width, height, left, top, longestSide;
   double dist;
      
   if (x1<x2) {
      left = x1; width = x2-x1;
   } else {
      left = x2; width = x1-x2;
   }
   if (y_1<y_2) {
      top = y_1; height = y_2-y_1;
   } else {
      top = y_2; height = y_1-y_2;
   }
   dist = (double)sqrt( (double)((x2-x1)*(x2-x1) + (y_2-y_1)*(y_2-y_1)) );
   longestSide = (width>height)?width:height;
      
   switch (dragVisual) {
      case MPE_DRAG_LINE:
         *cx1 = x1;
         *cy_1 = y_1;
         *cwidth = x2;
         *cheight = y_2;
         break;
            
      case MPE_DRAG_FIXED_RECT:
         if (width*ratio > height) {
            /* width is limiting factor */
            height = width * ratio;
         } else {
            width = height / ratio;
         }
         *cx1 = (x1>x2) ? (x1-width)  : x1;
         *cy_1 = (y_1>y_2) ? (y_1-height) : y_1;
         *cwidth = width;
         *cheight = height;
         break;
            
      default:
      case MPE_DRAG_RECT:
            
         *cx1 = left;
         *cy_1 = top;
         *cwidth = width;
         *cheight = height;
         break;
   }
      
   return 0;
}
   
void display_set_colorscale(double scale_in){
   colScale = scale_in;   

   if(graScale<colScale) graScale=colScale;
}

void display_set_graphscale(double scale_in){
   graScale = scale_in;   

   if(graScale<colScale) graScale=colScale;
}

void display_set_graphrow(unsigned int row_in){
   graRow = row_in;   
}
   
void display_set_window(double xmin_in, double xmax_in, double ymin_in, double ymax_in){
   xmin = xmin_in;
   xmax = xmax_in;
   ymin = ymin_in;
   ymax = ymax_in;
   
   xconv = (double)conWidth/ (xmax-xmin);
   yconv = (double)conHeight/(ymax-ymin);
}
   
#ifdef XXX
   glFlush();
   glutSwapBuffers();
   glClearColor(0.0, 0.0, 0.0, 0.0);
   glClear(GL_COLOR_BUFFER_BIT);
   glClear(GL_DEPTH_BUFFER_BIT);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glFlush();
#endif
