/*
 *  mesh.c
 *
 *  Copyright 2007-2010 Jon Robey and Bob Robey, brobey@lanl.gov 
 *  All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 */
#include "config.h"

#ifdef USE_MPI
#include <mpi.h>
#endif
#include <stdio.h>
#include <math.h>
#include "mesh.h"
#include "display.h"
#include "memory.h"

static unsigned int isize    = -1;
static unsigned int jsize    = -1;
#ifdef USE_MPI
int numpes   = -1;
int rank     = -1;
static unsigned int prev     = MPI_PROC_NULL;
static unsigned int next     = MPI_PROC_NULL;
#else
int numpes=1;
int rank=0;
#endif
static unsigned int mysize   = -1;
static unsigned int myoffset = -1;
static unsigned int nbound   =  2; // Needed for 2nd order method
static unsigned int debug    =  0;
static double deltaX = 0.0;
static double deltaY = 0.0;
static double xmin = 0.0;
static double ymin = 0.0;
static double xmax = 100.0;
static double ymax = 100.0;

struct esum_type{
    double sum;
    double correction;
};

#ifdef USE_MPI
MPI_Datatype MPI_TWO_DOUBLES;
MPI_Op KAHAN_SUM;
void kahan_sum(struct esum_type *in, struct esum_type *inout, int *len, MPI_Datatype *MPI_TWO_DOUBLES);
#define COMMUTATIVE 1
#endif

void mesh_set_size(unsigned int imax, unsigned int jmax)
{
  isize = imax;
  jsize = jmax;

  /* Determine numpes and my rank in MPI_COMM_WORLD communicator */
#ifdef USE_MPI
  MPI_Comm_size(MPI_COMM_WORLD, &numpes) ;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank) ;

  if (rank != 0)         prev = rank-1;
  if (rank < numpes - 1) next = rank+1;
#endif

  mysize = jmax/numpes + ((rank < (jmax % numpes)) ? 1 : 0 ) ;

  myoffset = rank * (jmax/numpes);
  if (rank > (jmax % numpes)) myoffset += (jmax % numpes);
  else                        myoffset += rank;
}

unsigned int mesh_get_mysize(void) { return(mysize); }
unsigned int mesh_get_myoffset(void) { return(myoffset); }

void mesh_initialize(unsigned int problem, double **H, double **U, double **V)
{
  if (problem == SLOPED_WAVE){
    for(unsigned int j=0; j<mysize+2*nbound; j++){
      for(unsigned int i=0; i<isize+2*nbound; i++){
        H[j][i]=2.0;
        U[j][i]=0.0;
        V[j][i]=0.0;
      }
        for(unsigned int i=2; i<=isize/2+1; i++){
          H[j][i]=10.0 - ((10.0 - 2.0)/ (double)(isize/2))*(double)(i-1);
        }
    }
    deltaX=1.0;
    deltaY=1.0;
  }
  if (problem == WAVE_BOX){
    for(unsigned int j=0; j<mysize+2*nbound; j++){
      for(unsigned int i=0; i<isize+2*nbound; i++){
        H[j][i]=1.0;
        U[j][i]=0.0;
        V[j][i]=0.0;
        if(j+myoffset >= jsize/2 -jsize/75 -1 && j+myoffset <= jsize/2 + jsize/75 -1 &&
          i >= isize/2 -isize/75 -1 && i <= isize/2 +isize/75 -1 ){
            H[j][i]=10.0;
        }
      }
    }
    deltaX=1.0;
    deltaY=1.0;
  }
  xmax = xmin + ((double)isize+1)*deltaX;
  ymax = ymin + ((double)jsize+1)*deltaY;
}

double mesh_get_deltaX(void) { return(deltaX); }
double mesh_get_deltaY(void) { return(deltaY); }
double mesh_get_xmin(void) { return(xmin); }
double mesh_get_ymin(void) { return(ymin); }
double mesh_get_xmax(void) { return(xmax); }
double mesh_get_ymax(void) { return(ymax); }

double *mesh_get_x(void){
  double *x=dvector(isize);
  for (int i=0; i<isize; i++){
    x[i] = xmin + ((double)i+0.5)*deltaX;
  }
  return(x);
}

double *mesh_get_dx(void){
  double *dx=dvector(isize);
  for (int i=0; i<isize; i++){
    dx[i] = deltaX;
  }
  return(dx);
}

double *mesh_get_y(void){
  double *y=dvector(jsize);
  for (int j=0; j<jsize; j++){
    y[j] = ymin + ((double)j+0.5)*deltaY;
  }
  return(y);
}

double *mesh_get_dy(void){
  double *dy=dvector(jsize);
  for (int j=0; j<jsize; j++){
    dy[j] = deltaY;
  }
  return(dy);
}

void mesh_set_boundary_conditions(double **H, double **U, double **V)
{
  unsigned int length = isize+2*nbound;

#ifdef USE_MPI
  MPI_Request      req[24];
  MPI_Status       status[24];

  /* Send and receive boundary information */
  MPI_Isend(H[mysize],length,MPI_DOUBLE,next,1,MPI_COMM_WORLD,req);
  MPI_Irecv(H[0],length,MPI_DOUBLE,prev,1,MPI_COMM_WORLD,req+1);

  MPI_Isend(H[mysize+1],length,MPI_DOUBLE,next,2,MPI_COMM_WORLD,req+2);
  MPI_Irecv(H[1],length,MPI_DOUBLE,prev,2,MPI_COMM_WORLD,req+3);

  MPI_Isend(H[2],length,MPI_DOUBLE,prev,3,MPI_COMM_WORLD,req+4);
  MPI_Irecv(H[mysize+2],length,MPI_DOUBLE,next,3,MPI_COMM_WORLD,req+5);

  MPI_Isend(H[3],length,MPI_DOUBLE,prev,4,MPI_COMM_WORLD,req+6);
  MPI_Irecv(H[mysize+3],length,MPI_DOUBLE,next,4,MPI_COMM_WORLD,req+7);

  if(rank==0&&debug>=1){printf("values for H communicated\n");}

  MPI_Isend(U[mysize],length,MPI_DOUBLE,next,5,MPI_COMM_WORLD,req+8);
  MPI_Irecv(U[0],length,MPI_DOUBLE,prev,5,MPI_COMM_WORLD,req+9);

  MPI_Isend(U[mysize+1],length,MPI_DOUBLE,next,6,MPI_COMM_WORLD,req+10);
  MPI_Irecv(U[1],length,MPI_DOUBLE,prev,6,MPI_COMM_WORLD,req+11);

  MPI_Isend(U[2],length,MPI_DOUBLE,prev,7,MPI_COMM_WORLD,req+12);
  MPI_Irecv(U[mysize+2],length,MPI_DOUBLE,next,7,MPI_COMM_WORLD,req+13);

  MPI_Isend(U[3],length,MPI_DOUBLE,prev,8,MPI_COMM_WORLD,req+14);
  MPI_Irecv(U[mysize+3],length,MPI_DOUBLE,next,8,MPI_COMM_WORLD,req+15);

  if(rank==0&&debug>=1){printf("values for x momentum communicated\n");}

  MPI_Isend(V[mysize],length,MPI_DOUBLE,next,9,MPI_COMM_WORLD,req+16);
  MPI_Irecv(V[0],length,MPI_DOUBLE,prev,9,MPI_COMM_WORLD,req+17);

  MPI_Isend(V[mysize+1],length,MPI_DOUBLE,next,10,MPI_COMM_WORLD,req+18);
  MPI_Irecv(V[1],length,MPI_DOUBLE,prev,10,MPI_COMM_WORLD,req+19);

  MPI_Isend(V[2],length,MPI_DOUBLE,prev,11,MPI_COMM_WORLD,req+20);
  MPI_Irecv(V[mysize+2],length,MPI_DOUBLE,next,11,MPI_COMM_WORLD,req+21);

  MPI_Isend(V[3],length,MPI_DOUBLE,prev,12,MPI_COMM_WORLD,req+22);
  MPI_Irecv(V[mysize+3],length,MPI_DOUBLE,next,12,MPI_COMM_WORLD,req+23);

  if(rank==0&&debug>=1){printf("values for y momentum communicated\n");}
  MPI_Waitall(24, req, status);
#endif

  if(rank==0&&debug>=1)printf("Communication successful\n");
  //set boundary conditions
  for(unsigned int j=nbound;j<mysize+nbound;j++){
    H[j][0]= H[j][2];
    U[j][0]=-U[j][2];
    V[j][0]= V[j][2];
    H[j][1]= H[j][2];
    U[j][1]=-U[j][2];
    V[j][1]= V[j][2];
    H[j][isize+2]= H[j][isize+1];
    U[j][isize+2]=-U[j][isize+1];
    V[j][isize+2]= V[j][isize+1];
    H[j][isize+3]= H[j][isize+1];
    U[j][isize+3]=-U[j][isize+1];
    V[j][isize+3]= V[j][isize+1];
    }
  for(unsigned int i=0;i<isize+nbound;i++){
    if(myoffset==0){
      H[0][i]= H[2][i];
      U[0][i]= U[2][i];
      V[0][i]=-V[2][i];
      H[1][i]= H[2][i];
      U[1][i]= U[2][i];
      V[1][i]=-V[2][i];
    }
    if(jsize==myoffset+mysize){
      H[mysize+2][i]= H[mysize+1][i];
      U[mysize+2][i]= U[mysize+1][i];
      V[mysize+2][i]=-V[mysize+1][i];
      H[mysize+3][i]= H[mysize+1][i];
      U[mysize+3][i]= U[mysize+1][i];
      V[mysize+3][i]=-V[mysize+1][i];
    }
  }
}

double mesh_set_timestep(double g, double sigma,
    double **H, double **U, double **V)
{
  double wavespeed, xhold, yhold, xspeedmax, yspeedmax, deltaT;

  double xlocalmax=0;
  double ylocalmax=0;

  for(unsigned int j=nbound; j<mysize+nbound; j++){
    for(unsigned int i=nbound; i<isize+nbound; i++){
      wavespeed = sqrt(g*H[j][i]);
      xhold=(fabs(U[j][i])+wavespeed);
      if(xhold>xlocalmax) xlocalmax=xhold;
      yhold=(fabs(V[j][i])+wavespeed);
      if(yhold>ylocalmax) ylocalmax=yhold;
    }
  }

#ifdef USE_MPI
  MPI_Allreduce(&xlocalmax, &xspeedmax, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
  MPI_Allreduce(&ylocalmax, &yspeedmax, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
#else
  xspeedmax = xlocalmax;
  yspeedmax = ylocalmax;
#endif

  deltaT=sigma/(xspeedmax/deltaX+yspeedmax/deltaY);

  return(deltaT);
}

void mesh_output(char *title, unsigned int imin, unsigned int imax,
    unsigned int jmin, unsigned int jmax, double **H, double **U, double **V)
{
    //Should put allgather in here to print all the mesh data

    printf("\n   %s\n\n",title);
    for(unsigned int j=jmin; j<jmax; j++){
      for(unsigned int i=imin; i<imax; i++){
        printf("%s -- i %d j %d H %lf U %lf V %lf\n",title, i, j, H[j][i],
            U[j][i], V[j][i]);
      }
    }
}

void cycle_print(double **plot_var, unsigned int n, unsigned int icount, double time,
    double deltaT, double TotalMass, double **mesh_var, double colorScale,
    double graphScale, unsigned int meshRow, char *label)
{
  if (n%icount == 0){
    if (rank == 0) {
      printf("Iteration:%5.5d, Time:%f, Timestep:%f Total mass:%.10f \n",
          n, time, deltaT, TotalMass);
    }
  }
}

#ifdef USE_MPI
void sum_init(void){
  MPI_Type_contiguous(2, MPI_DOUBLE, &MPI_TWO_DOUBLES);
  MPI_Type_commit(&MPI_TWO_DOUBLES);
  MPI_Op_create((MPI_User_function *)kahan_sum, COMMUTATIVE, &KAHAN_SUM);
}

void sum_finalize(void){
  MPI_Op_free(&KAHAN_SUM);
  MPI_Type_free(&MPI_TWO_DOUBLES);
}
#endif

double sum_var(double **var)
{
  double corrected_next_term, new_sum;
  struct esum_type local, global;

  // Robey, Robey and Aulwes -- In Search of Numerical Consistency in Parallel
  //    Programming
  // Kahan summation for more precision -- see Wikipedia or Dr. Dobbs article

  local.sum=0.0;
  local.correction=0.0;
  for(unsigned int j=nbound; j<mysize+nbound; j++){
    for(unsigned int i=nbound; i<isize+nbound; i++){
        corrected_next_term = var[j][i]*deltaX*deltaY + local.correction;
        new_sum = local.sum + corrected_next_term;
        local.correction = corrected_next_term - (new_sum - local.sum);
        local.sum = new_sum;
    }
  }

#ifdef USE_MPI
  MPI_Allreduce(&local, &global, 1, MPI_TWO_DOUBLES, KAHAN_SUM, MPI_COMM_WORLD);
#else
  global.sum = local.sum;
#endif

  return(global.sum);
}

#ifdef USE_MPI
void kahan_sum(struct esum_type *in, struct esum_type *inout, int *len, MPI_Datatype *MPI_TWO_DOUBLES)
{
  double corrected_next_term, new_sum;

  corrected_next_term = in->sum + (in->correction+inout->correction);
  new_sum = inout->sum + corrected_next_term;
  inout->correction = corrected_next_term - (new_sum - inout->sum);
  inout->sum = new_sum;
}
#endif

void mesh_set_plot_var(double **mesh_var, double **plot_var)
{
  for(unsigned int j=0; j<mysize; j++){
    for(unsigned int i=0; i<isize; i++){
      plot_var[j][i]=mesh_var[j+nbound][i+nbound];
    }
  }
}
